from mock import Mock
from munch import munchify as m

from s3manage.buckets import Buckets


def test_buckets_work():
    raw = [m({'name': 'valid'}), m({'name': 'valid2'}), m({'name': '1'})]
    res = Mock(**{'buckets.all.return_value': raw})
    buckets = Buckets(res, Mock())
    all_buckets = buckets.get_buckets()
    assert len(all_buckets) == 3
    assert all_buckets[0].get_name() == 'valid'
    assert all_buckets[1].get_name() == 'valid2'
    assert all_buckets[2].get_name() == '1'
    bucket = buckets.get_bucket('valid2')
    assert bucket.get_name() == 'valid2'
    bucket = buckets.get_bucket('nope')
    assert bucket is None
    valid, invalid = buckets.validate_buckets(['valid', '1', 'invalid'])
    assert len(valid) == 2
    assert len(invalid) == 1


def test_empty_buckets():
    res = Mock(**{'buckets.all.return_value': []})
    buckets = Buckets(res, Mock())
    all_buckets = buckets.get_buckets()
    assert len(all_buckets) == 0
    valid, invalid = buckets.validate_buckets(['test', 'test1'])
    assert len(valid) == 0
    assert len(invalid) == 2
    bucket = buckets.get_bucket('nope')
    assert bucket is None



# -*- coding: UTF-8 -*-
from __future__ import unicode_literals

import mock
import pytest
import six
from munch import munchify

from s3manage.index import HTML_BASE
from s3manage.index import DirNode
from s3manage.index import Node
from s3manage.index import NodeError
from s3manage.index import RootNode
from s3manage.index import VirtualNode
from s3manage.index import node_factory


def test_root_node_only():
    node = node_factory(munchify({'key': ''}))
    assert node.is_dir
    assert node._item.key == ''
    assert node.get_index_file() == ROOT_NODE_ONLY_INDEX
    assert node.get_part(7) is None

def test_dir_node_only():
    node = DirNode(munchify({'key': 'asdf/', 'size': 0, 'last_modified': '2022-01-77'}))
    assert node.is_dir
    assert len(node.parts) == 1
    assert node.get_index_file() == DIR_NODE_ONLY_INDEX
    str(node)

def test_first_level_children(capsys):
    node = RootNode()
    node.add_child(Node(munchify({'key': 'derp.txt', 'size': 177, 'last_modified': '2017-01-01'})))
    node.add_child(Node(munchify({'key': 'asdf/herp.dat', 'size': 10000, 'last_modified': '2016-01-02'})))
    node.add_child(DirNode(munchify({'key': 'asdf/', 'size': 0, 'last_modified': '2016-01-03'})))
    assert node.is_dir
    assert len(node._children) == 2
    assert not node._children['derp.txt'].is_dir
    assert node._children['asdf'].is_dir
    assert len(node._children['asdf']._children) == 1
    assert node.get_index_file() == FIRST_LEVEL_CHILDREN_INDEX
    assert node._children['asdf'].get_index_file() == FIRST_LEVEL_CHILDREN_L2_INDEX
    node.print_tree()
    out, err = capsys.readouterr()
    assert out == FIRST_LEVEL_CHILDREN_TREE
    dest = mock.Mock()
    node.write_index_file(dest, 'index.html')
    first = mock.call(ContentType='text/html',
                 Body=bytearray(FIRST_LEVEL_CHILDREN_INDEX, 'utf-8'),
                 Key='index.html')
    second = mock.call(ContentType='text/html',
                  Body=bytearray(FIRST_LEVEL_CHILDREN_L2_INDEX, 'utf-8'),
                  Key='asdf/index.html')
    dest.put_object.assert_has_calls([first, second])
    str(node._children['derp.txt'])

def test_basic_node():
    node = Node(munchify({'key': 'asdf/derp.txt'}))
    assert node.get_index_file() is None
    links = []
    node.get_index_file_links(links)
    assert len(links) == 0
    assert not node.is_dir
    # On a standard node, this method should not operate on the destination at
    # all
    destination = mock.Mock()
    node.write_index_file(destination, 'index.html')
    assert len(destination.method_calls) == 0

def test_virtual_node():
    node = VirtualNode('testdir/')
    assert node.is_dir
    assert node.virtual
    assert len(node.parts) == 1

def test_non_virtual_node_complains():
    node = Node(munchify({'key': 'repeat'}))
    with pytest.raises(NodeError) as e_info:
        node.set_node(node)

ROOT_NODE_ONLY_INDEX=HTML_BASE.format("")

DIR_NODE_ONLY_INDEX=HTML_BASE.format('<tr><td></td> <td><a href="..">&lt;Parent&gt;</a></td><td></td><td></td></tr>')

FIRST_LEVEL_CHILDREN_INDEX=HTML_BASE.format("""<tr><td>📁</td><td><a href="asdf">asdf</a></td><td>2016-01-03</td><td> - </td></tr>
<tr><td>📃</td><td><a href="derp.txt">derp.txt</a></td><td>2017-01-01</td><td>177 bytes</td></tr>""")

FIRST_LEVEL_CHILDREN_L2_INDEX=HTML_BASE.format("""<tr><td></td> <td><a href="..">&lt;Parent&gt;</a></td><td></td><td></td></tr>
<tr><td>📃</td><td><a href="herp.dat">herp.dat</a></td><td>2016-01-02</td><td>10 KB</td></tr>""")

FIRST_LEVEL_CHILDREN_TREE="""|--- 
|   |--- asdf/
|   |   |--- asdf/herp.dat
|   |--- derp.txt
"""



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -ms3manage` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``s3manage.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``s3manage.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
from __future__ import print_function

import argparse
import sys

from s3manage.buckets import Buckets
from s3manage.index import Index

parser = argparse.ArgumentParser(description='Command description.')
parser.add_argument('buckets', metavar='NAME', nargs=argparse.ONE_OR_MORE,
                    help="The bucket to add indexes to")
parser.add_argument('--include-indexes', action='store_true',
                    help="Include the index files in the generated files")
parser.add_argument('--base-path', '-b', metavar='PATH', action='store',
                    default='',
                    help="Only generate index files under the specified path")


def indexes(args=None):
    args = parser.parse_args(args=args)
    print("Connecting to S3")
    buckets = Buckets()
    print("Verifying buckets")
    valid, invalid = buckets.validate_buckets(args.buckets)
    # Buckets we cannot find indicate a problem
    if invalid:
        print("Missing buckets: ")
        [print("    {0}".format(name)) for name in invalid]
        sys.exit(1)
    for bucket in valid:
        index = Index(bucket)
        print("Constructing index for {0}".format(bucket.get_name()))
        index.construct()
        index.write_indexes()



# -*- coding: UTF-8 -*-
from __future__ import unicode_literals

import mock
from munch import munchify as m

from s3manage.index import HTML_BASE
from s3manage.index import Index

test_data = [mock.Mock(**{'key': 'test.txt', 'last_modified': '2011-01-01', 'size': 100000000}),
             mock.Mock(**{'key': 'help.dat', 'last_modified': '2011-01-01', 'size': 100000000}),
             mock.Mock(**{'key': 'firstdir/', 'last_modified': '2011-01-01', 'size': 0}),
             mock.Mock(**{'key': 'firstdir/firstfile.txt', 'last_modified': '2011-01-01', 'size': 100000000}),
             mock.Mock(**{'key': 'firstdir/nope.exe', 'last_modified': '2011-01-01', 'size': 100000000}),
             mock.Mock(**{'key': 'seconddir/', 'last_modified': '2011-01-01', 'size': 0}),
             mock.Mock(**{'key': 'notadir', 'last_modified': '2011-01-01', 'size': 100000000})]


def test_create_index():
    mock_bucket = mock.Mock(**{'get_items.return_value': test_data,
                               'get_index_name.return_value': 'index.html'})
    index = Index(mock_bucket)
    index.construct()
    assert index._root.is_dir
    assert len(index._root._children) == 5
    assert len(index._root._children['firstdir']._children) == 2
    assert len(index._root._children['seconddir']._children) == 0
    assert not index._root._children['notadir'].is_dir
    index.write_indexes()
    c1 = mock.call(ContentType="text/html",
              Body=bytearray(FIRST_LEVEL, 'utf-8'),
              Key="index.html")
    c2 = mock.call(ContentType="text/html",
              Body=bytearray(SECOND_LEVEL, 'utf-8'),
              Key="firstdir/index.html")
    c3 = mock.call(ContentType="text/html",
              Body=bytearray(SECOND_LEVEL_2, 'utf-8'),
              Key="seconddir/index.html")
    mock_bucket.put_object.assert_has_calls([c1, c2, c3])


FIRST_LEVEL=HTML_BASE.format("""<tr><td>📁</td><td><a href="firstdir">firstdir</a></td><td>2011-01-01</td><td> - </td></tr>
<tr><td>📃</td><td><a href="help.dat">help.dat</a></td><td>2011-01-01</td><td>100 MB</td></tr>
<tr><td>📃</td><td><a href="notadir">notadir</a></td><td>2011-01-01</td><td>100 MB</td></tr>
<tr><td>📁</td><td><a href="seconddir">seconddir</a></td><td>2011-01-01</td><td> - </td></tr>
<tr><td>📃</td><td><a href="test.txt">test.txt</a></td><td>2011-01-01</td><td>100 MB</td></tr>""")

SECOND_LEVEL=HTML_BASE.format("""<tr><td></td> <td><a href="..">&lt;Parent&gt;</a></td><td></td><td></td></tr>
<tr><td>📃</td><td><a href="firstfile.txt">firstfile.txt</a></td><td>2011-01-01</td><td>100 MB</td></tr>
<tr><td>📃</td><td><a href="nope.exe">nope.exe</a></td><td>2011-01-01</td><td>100 MB</td></tr>""")

SECOND_LEVEL_2=HTML_BASE.format('<tr><td></td> <td><a href="..">&lt;Parent&gt;</a></td><td></td><td></td></tr>')



# -*- coding: UTF-8 -*-
from __future__ import print_function
from __future__ import unicode_literals

import posixpath as webpath
from builtins import str

import humanfriendly
import six
from munch import munchify


HTML_BASE = """<!DOCTYPE html>
<html>
    <head>
        <meta charset="UTF-8" />
        <style type="text/css">
            thead {{
                text-align: center;
                font-weight: bold;
                font-size: 125%;
            }}
            thead th {{
                padding-bottom: 10px;
            }}
            tbody {{
                border-top: solid 1px #aaa;
                border-bottom: solid 1px #aaa;
            }}
            tbody td {{
                padding-right: 5px;
            }}
            tbody tr:hover {{
                background-color: #ddd;
                border: 1px solid #aaa;
            }}
        </style>
    </head>
    <body>
        <table>
            <thead>
                <tr>
                    <th></th>
                    <th>Name</th>
                    <th>Last Modified</th>
                    <th>Size</th>
                </tr>
            </thead>
            <tbody>
{0}
            </tbody>
        </table>
    </body>
</html>"""


class Index(object):
    """
    Models and can optionally create a set of static HTML files that link to
    the given S3 bucket's files in the vein of Apache's +Indexes option.
    """
    def __init__(self, bucket):
        self._bucket = bucket

    def construct(self):
        files = self._bucket.get_items()
        # Find the root
        self._root = RootNode()
        # Construct the tree, in memory
        for f in files:
            node = node_factory(f)
            self._root.add_child(node)

    def write_indexes(self):
        self._root.write_index_file(self._bucket, self._bucket.get_index_name())


def node_factory(item):
    if item.key == '':
        return RootNode()
    elif item.key.endswith('/'):
        return DirNode(item)
    else:
        return Node(item)


class Node(object):
    """
    A base node, anywhere in the tree. Behanves mostly as a single, stand-alone
    item and doesn't take too much stock in its cihldren, although it can have
    some. It's more or less like a deadbeat parent who just has kids and then
    ignores them unless it's very convenient to itself.

    Use as the base class for other types of nodes.
    """
    def __init__(self, item):
        self._item = item
        self._children = {}
        self.virtual = False
        self.is_dir = self._item.key.endswith('/')
        self.parts = self._item.key.split('/')
        self.icon = '📃'

    def __str__(self):
        if six.PY2:
            return self.__unicode__().encode('utf-8')
        else:
            return self.__unicode__()

    def __unicode__(self):
        rep = '<tr><td>{0}</td>'.format(self.icon)
        rep += '<td><a href="{0}">{0}</a></td>'.format(self.parts[-1])
        rep += '<td>{0}</td>'.format(self._item.last_modified)
        size = humanfriendly.format_size(self._item.size) if self._item.size > 0 else ' - '
        rep += '<td>{0}</td>'.format(size)
        rep += '</tr>'
        return rep

    def add_child(self, node, depth=0):
        """
        Adds a child node into the hierarchy. The child node should already be
        instantiated as part of itself.

        :param node: The node that is going to be inserted into this hierarchy
        :param depth: The depth down the tree that this node should descend.
        Strictly speaking, this parameter might be removed in the future, so do
        not depend on it from outside of this class itself
        """
        part = node.get_part(depth)
        # If new node is at its max depth, end recursion.
        if depth == len(node.parts) - 1:
            if part in self._children:
                self._children[part].set_node(node)
            else:
                self._children[part] = node
            return
        # Create a virtual node, if we don't have an existing node for it
        if part not in self._children:
            s = '/'.join(node.parts[:depth+1]) + '/'
            self._children[part] = VirtualNode(s)
        # Add this incoming child to the trie
        self._children[part].add_child(node, depth+1)

    def get_part(self, part):
        """
        Get a path segment at the specified depth in this node's tree structure

        :param part: Index depth down the path structure to fetch back. 0-index
        :returns: The path segment at the specified depth, None if the request
        is out of bounds
        """
        if part > len(self.parts):
            return None
        else:
            return self.parts[part]

    def set_node(self, other):
        """
        Replace myself with the other node that has been discovered. The main
        purpose for this method is to transform a formerly virtual node into a
        concrete node. Initializes all the variables appropriately.

        :param other: The concrete node with which to replace our own values
        """
        if not self.virtual:
            raise NodeError('wtf: You have the same thing twice?')
        self._item = other._item
        self.parts = other.parts
        self.is_dir = other.is_dir
        self.virtual = False

    def print_tree(self, indent=0):
        if indent > 0:
            print("|   " * indent, end='')
        print("|---", self._item.key)

    def get_index_file_links(self, links):
        pass

    def get_index_file(self):
        return None

    def write_index_file(self, destination, name):
        pass


class DirNode(Node):
    """
    A node that is speficially analgous to a filesystem directory or folder.
    Unlike a generic node, this one pays much closer attention to its children
    and works with them as often as they can appear.
    """
    def __init__(self, item):
        super(DirNode, self).__init__(item)
        self.is_dir = True
        self.parts = self.parts[:-1]
        self.icon = '📁'

    def __str__(self):
        if six.PY2:
            return self.__unicode__().encode('utf-8')
        else:
            return self.__unicode__()

    def get_index_file(self):
        links = ['<tr><td></td> <td><a href="..">&lt;Parent&gt;</a></td><td></td><td></td></tr>']
        self.get_index_file_links(links)
        return self.add_links_to_body(links)

    def add_links_to_body(self, links):
        # Construct the full file
        link_strs = '\n'.join(links)
        f = HTML_BASE.format(link_strs)
        return f

    def get_index_file_links(self, links):
        # Create each link for child elements
        for child in sorted(self._children.keys()):
            links.append(str(self._children[child]))

    def print_tree(self, indent=0):
        super(DirNode, self).print_tree(indent)
        for child in sorted(self._children.keys()):
            self._children[child].print_tree(indent+1)

    def write_index_file(self, destination, name):
        mine = self.get_index_file()
        key = webpath.join(self._item.key, name)
        destination.put_object(ContentType='text/html',
                               Body=bytearray(mine, 'utf-8'),
                               Key=key)
        for child in sorted(self._children.keys()):
            self._children[child].write_index_file(destination, name)


class VirtualNode(DirNode):
    """
    A node that, strictly speaking, should be present in the tree but, for
    whatever reason, does not yet exist there. This could be because it was
    created after elements that are contained within itself. Or it could be
    because the S3 API might allow the direct creation of elements without
    explicitly creating their folder structure.
    """
    def __init__(self, path):
        o = munchify({'key': path})
        super(VirtualNode, self).__init__(o)
        self.virtual = True


class RootNode(DirNode):
    def __init__(self):
        super(RootNode, self).__init__(munchify({'key': ''}))
        self.virtual = True

    def get_index_file(self):
        links = []
        self.get_index_file_links(links)
        return self.add_links_to_body(links)


class NodeError(Exception):
    pass



from __future__ import print_function

import boto3


class Buckets(object):
    """
    Represents an Amazon S3 account, and provides operations on the buckets in
    various ways.
    """
    def __init__(self,
                 res=boto3.resource('s3'),
                 client=boto3.client('s3')):
        self._res = res
        self._client = client
        self._buckets = [Bucket(self._client, bucket) for bucket in
                         self._res.buckets.all()]

    def get_buckets(self):
        """
        Fetch back all of the s3 buckets, wrapped in our own helper class.

        :returns: Array of all buckets
        """
        return self._buckets

    def get_bucket(self, name):
        """
        Get a single wrapper object for the bucket with the given name.

        :return: A Bucket of the given name. None if no such bucket exists.
        """
        for bucket in self.get_buckets():
            if bucket.get_name() == name:
                return bucket
        return None

    def validate_buckets(self, buckets):
        '''
        Tests to be sure that all the buckets in the provided list are
        available on the s3 client specified.

        :param buckets: An iterable of names to check for bucket status
        :returns: (valid, invalid) tuple of iterables of buckets that are valid
        vs names that are invalid
        '''
        # This is 2*n^2, but it makes for prettier code. If your number of
        # buckets is really that high, then we can discuss making this more
        # n^2 or even look at a linear option
        present = [self.get_bucket(b) for b in buckets if self.get_bucket(b)]
        missing = [b for b in buckets if not self.get_bucket(b)]
        return present, missing


class Bucket(object):
    """
    Represents a single bucket, and wraps some functionality not provided by
    the underlying bucket object.
    """
    def __init__(self, client, bucket):
        self._client = client
        self._bucket = bucket

    def get_name(self):
        """
        Get the name of the bucket

        :returns: The string name of the bucket, as reported by S3
        """
        return self._bucket.name

    def get_index_name(self):
        """
        Gets the name of the configured Index file for this bucket

        :returns: The name of the index file for this bucket
        """
        config = self._client.get_bucket_website(Bucket=self.get_name())
        if 'IndexDocument' not in config:
            print("Bucket not configured for website, assuming " +
                  "'index.html'")
            return 'index.html'
        else:
            return config['IndexDocument']['Suffix']

    def get_items(self):
        """
        Fetch a listing of all the items in this bucket.

        :return: Array of objects in this bucket. The 'key' element on each
        item is the name of that object.
        """
        return [o for o in self._bucket.objects.all()]

    def put_object(self, **kargs):
        self._bucket.put_object(**kargs)



from mock import Mock

from s3manage.buckets import Bucket

MOCK_NAME = "Test Bucket"
index = {'IndexDocument': {'Suffix': 'my_index.html'}}
raw_client = Mock(**{'get_bucket_website.return_value': index})
raw_bucket = Mock(**{'objects.all.return_value' : ['1', '2']})
raw_bucket.name = MOCK_NAME


def test_bucket():
    bucket = Bucket(raw_client, raw_bucket)
    items = bucket.get_items()
    assert bucket.get_name() == MOCK_NAME
    assert len(items) == 2
    assert bucket.get_index_name() == index['IndexDocument']['Suffix']
    args = {'something': 123, 'another': 'xyz'}
    bucket.put_object(**args)
    raw_bucket.put_object.assert_called_with(**args)


def test_bucket_no_website():
    client = Mock(**{'get_bucket_website.return_value': {}})
    bucket = Bucket(client, raw_bucket)
    assert bucket.get_index_name() == 'index.html'


def test_bucket_empty():
    empty_bucket = Mock(**{'name': MOCK_NAME, 'objects.all.return_value': []})
    bucket = Bucket(raw_client, empty_bucket)
    assert len(bucket.get_items()) == 0
